import numpy as np


class Ogden:
    def __init__(self, mu, alpha):
        self.mu = mu
        self.alpha = alpha

    def gradient(self, stretches, statevars):
        dWdλ = 2 * self.mu / self.alpha * stretches ** (self.alpha - 1)
        return dWdλ, statevars

    def hessian(self, stretches, statevars):
        d2Wdλdλ = (
            2 * self.mu / self.alpha * (self.alpha - 1) * stretches ** (self.alpha - 2)
        )
        return np.concatenate([d2Wdλdλ, np.zeros_like(d2Wdλdλ)])



import numpy as np

from ..math import cdya, ddot, dya, eye, trace


class InvariantsFramework:
    r"""The Framework for a Total-Lagrangian invariant-based isotropic hyperelastic
    material formulation provides the material behaviour-independent parts for
    evaluating the second Piola-Kirchhoff stress tensor as well as its associated
    fourth-order elasticity tensor.

    The gradient as well as the hessian of the strain energy function are carried out
    w.r.t. the right Cauchy-Green deformation tensor. Hence, the work-conjugate stress
    tensor is one half of the second Piola-Kirchhoff stress tensor and the fourth-order
    elasticitiy tensor used here is a quarter of the Total-Lagrangian elasticity tensor.

    ..  math::

        \psi(\boldsymbol{C}) =
            \psi(I_1(\boldsymbol{C}), I_2(\boldsymbol{C}), I_3(\boldsymbol{C}))

    The first and second invariants of the left or right Cauchy-Green deformation tensor
    are identified as factors of their characteristic polynomial,

    ..  math::

        I_1 &= \text{tr}(\boldsymbol{C})

        I_2 &= \frac{1}{2}
            \left( \text{tr}(\boldsymbol{C})^2 - \text{tr}(\boldsymbol{C}^2) \right)

    where the Cauchy-Green deformation tensors eliminate the rigid body rotations
    of the deformation gradient and serve as a quadratic change-of-length measure of
    the deformation.

    ..  math::

        \boldsymbol{C} &= \boldsymbol{F}^T \boldsymbol{F}

        \boldsymbol{b} &= \boldsymbol{F} \boldsymbol{F}^T

    The first partial derivatives of the strain energy function w.r.t. the invariants

    ..  math::

        \psi_{,1} &= \frac{\partial \psi}{\partial I_1}

        \psi_{,2} &= \frac{\partial \psi}{\partial I_2}

    and the partial derivatives of the invariants w.r.t. the right Cauchy-Green
    deformation tensor are defined.

    ..  math::

        \frac{\partial I_1}{\partial \boldsymbol{C}} &= 2 \boldsymbol{1}

        \frac{\partial I_2}{\partial \boldsymbol{C}} &=
            2 \left( I_1 \boldsymbol{1} - \boldsymbol{C} \right)

    The second Piola-Kirchhoff stress tensor is formulated by the application of the
    chain rule.

    ..  math::

        \frac{\partial \psi}{\partial \boldsymbol{C}} =
            \frac{\partial \psi}{\partial I_1}
            \frac{\partial I_1}{\partial \boldsymbol{C}}
            + \frac{\partial \psi}{\partial I_2}
            \frac{\partial I_2}{\partial \boldsymbol{C}}

    Furthermore, the second partial derivatives of the elasticity tensor are carried
    out.

    ..  math::

        \mathbb{C} &= \frac{\partial^2 \psi}{\partial I_1~\partial I_1}
            \left( \frac{\partial I_1}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_1}{\partial \boldsymbol{C}} \right)
            + \frac{\partial^2 \psi}{\partial I_2~\partial I_2}
            \left( \frac{\partial I_2}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_2}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial^2 \psi}{\partial I_1~\partial I_2}
            \left( \frac{\partial I_1}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_2}{\partial \boldsymbol{C}}
            + \frac{\partial I_2}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_1}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial \psi}{\partial I_1}
            \frac{\partial^2 I_1}{\partial \boldsymbol{C}~\partial \boldsymbol{C}}
            + \frac{\partial \psi}{\partial I_2}
            \frac{\partial^2 I_2}{\partial \boldsymbol{C}~\partial \boldsymbol{C}}

    The only non material behaviour-related term which is not already defined during
    stress evaluation is the second partial derivatives of the second invariant w.r.t.
    the right Cauchy-Green deformation tensor.

    ..  math::

        \frac{\partial^2 I_2}{\partial \boldsymbol{C}~\partial \boldsymbol{C}} &=
            \boldsymbol{I} \otimes \boldsymbol{I} - \boldsymbol{I} \odot \boldsymbol{I}

    """

    def __init__(self, material, nstatevars=0, parallel=False):
        """Initialize the Framework for an invariant-based isotropic hyperelastic
        material formulation."""

        self.parallel = parallel
        self.material = material

        self.x = [np.eye(3), np.zeros(nstatevars)]

    def gradient(self, C, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the right Cauchy-Green deformation tensor (one half of the second Piola
        Kirchhoff stress tensor)."""

        self.I = I = eye(C)

        self.I1 = I1 = trace(C)
        self.I2 = (I1**2 - ddot(C, C)) / 2

        self.dWdI1, self.dWdI2, statevars_new = self.material.gradient(
            self.I1, self.I2, statevars
        )

        self.dI1dC = I
        self.dI2dC = I1 * I - C

        dWdC = self.dWdI1 * self.dI1dC + self.dWdI2 * self.dI2dC

        return dWdC, statevars_new

    def hessian(self, C, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the right Cauchy-Green deformation tensor (a quarter of the Lagrangian
        fourth-order elasticity tensor associated to the second Piola-Kirchhoff stress
        tensor)."""

        dWdC, statevars = self.gradient(C, statevars)
        d2WdI1dI1, d2WdI2dI2, d2WdI1dI2 = self.material.hessian(
            self.I1, self.I2, statevars
        )
        I = self.I
        dI1dC = self.dI1dC
        dI2dC = self.dI2dC

        ntrax = len(C.shape[1:])
        d2WdCdC = np.zeros((6, 6, *np.ones(ntrax, dtype=int)))

        if not np.allclose(d2WdI1dI1, 0):
            d2WdCdC = d2WdCdC + d2WdI1dI1 * dya(dI1dC, dI1dC)

        if not np.allclose(d2WdI1dI2, 0):
            d2WdCdC = d2WdCdC + d2WdI1dI2 * (dya(dI1dC, dI2dC) + dya(dI2dC, dI1dC))

        if not np.allclose(self.dWdI2, 0):
            d2I2dCdC = dya(I, I) - cdya(I, I)

            d2WdCdC = d2WdCdC + self.dWdI2 * d2I2dCdC + d2WdI2dI2 * dya(dI2dC, dI2dC)

        return d2WdCdC



import numpy as np


def asvoigt(A, mode=2):
    """Convert 3x3 tensor to symmetric (Voigt-notation) vector storage of shape 6."""

    if mode == 2:  # 3x3 symmetric tensor
        i = [0, 1, 2, 0, 1, 0]
        j = [0, 1, 2, 1, 2, 2]

        return A[i, j]

    elif mode == 4:  # 3x3x3x3 major-symmetric tensor
        B = np.zeros((6, 6, *A.shape[4:]))
        for i in range(3):
            for j in range(3):
                B[i, j] = A[i, i, j, j]

        B[3, 3] = A[0, 1, 0, 1]
        B[4, 4] = A[1, 2, 1, 2]
        B[5, 5] = A[0, 2, 0, 2]

        for i in range(3):
            B[i, 3] = A[i, i, 0, 1]
            B[i, 4] = A[i, i, 1, 2]
            B[i, 5] = A[i, i, 0, 2]
            B[3, i] = A[0, 1, i, i]
            B[4, i] = A[1, 2, i, i]
            B[5, i] = A[0, 2, i, i]

        B[3, 4] = A[0, 1, 1, 2]
        B[3, 5] = A[0, 1, 0, 2]

        B[4, 3] = A[1, 2, 0, 1]
        B[4, 5] = A[1, 2, 0, 2]

        B[5, 3] = A[0, 2, 0, 1]
        B[5, 4] = A[0, 2, 1, 2]

        return B


def astensor(A, mode=2):
    "Convert symmetric 6 tensor in Voigt-notation to full 3x3 tensor."
    if mode == 2:  # second order tensor of shape 6 to 3x3
        a = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2]).reshape(3, 3)
        return A[a]

    elif mode == 4:  # fourth order tensor of shape 6x6 to 3x3x3x3
        a = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2])
        i, j = np.meshgrid(a, a)
        return A[i.reshape(3, 3, 3, 3), j.reshape(3, 3, 3, 3)]


def tril_from_triu(A, inplace=True):
    "Copy upper triangle values to lower triangle values of a 3x3 tensor inplace."
    B = A
    if not inplace:
        B = A.copy()

    i, j = np.triu_indices(6, 1)
    B[j, i] = A[i, j]
    return B


def triu_from_tril(A, inplace=True):
    "Copy lower triangle values to upper triangle values of a 3x3 tensor inplace."
    B = A
    if not inplace:
        B = A.copy()

    i, j = np.tril_indices(6, -1)
    B[j, i] = A[i, j]
    return B


def trace(A):
    "Return the sum of the diagonal values of a 3x3 tensor."
    return np.sum(A[:3], axis=0)


def transpose(A):
    "Return the major transpose of a fourth order tensor in Voigt-storage."
    return np.einsum("ij...->ji...", A)


def det(A):
    "The determinant of a symmetric 3x3 tensor in Voigt-storage (rule of Sarrus)."
    return (
        A[0] * A[1] * A[2]
        + 2 * A[3] * A[4] * A[5]
        - A[4] ** 2 * A[0]
        - A[5] ** 2 * A[1]
        - A[3] ** 2 * A[2]
    )


def inv(A, determinant=None):
    """The inverse of a symmetric 3x3 tensor in Voigt-storage with optional provided
    determinant."""

    detAinvA = np.zeros_like(A)

    if determinant is None:
        detA = det(A)
    else:
        detA = determinant

    detAinvA[0] = A[1] * A[2] - A[4] ** 2
    detAinvA[1] = A[0] * A[2] - A[5] ** 2
    detAinvA[2] = A[0] * A[1] - A[3] ** 2

    detAinvA[3] = A[4] * A[5] - A[3] * A[2]
    detAinvA[4] = A[3] * A[5] - A[0] * A[4]
    detAinvA[5] = A[3] * A[4] - A[1] * A[5]

    return detAinvA / detA


def dot(A, B, mode=(2, 2)):
    "The dot-product of two symmetric 3x3 tensors in Voigt-storage."

    if mode == (2, 2):
        trax = np.broadcast_shapes(A.shape[1:], B.shape[1:])
        C = np.zeros((3, 3, *trax))
        C[0] = A[0] * B[0] + A[3] * B[3] + A[5] * B[5]
        C[1] = A[3] * B[3] + A[1] * B[1] + A[4] * B[4]
        C[2] = A[4] * B[4] + A[5] * B[5] + A[2] * B[2]
        C[3] = A[0] * B[3] + A[3] * B[1] + A[5] * B[4]
        C[4] = A[3] * B[5] + A[1] * B[4] + A[4] * B[2]
        C[5] = A[0] * B[5] + A[3] * B[4] + A[5] * B[2]

    return C


def eye(A):
    "A 3x3 tensor in Voigt-storage with ones on the diagonal and zeros elsewhere."
    trax = np.ones(len(A.shape[1:]), dtype=int)
    return np.array([1, 1, 1, 0, 0, 0], dtype=float).reshape(6, *trax)


def ddot(A, B, mode=(2, 2)):
    "The double-contraction of two symmetric 3x3 tensors in Voigt-storage."
    weights = np.array([1, 1, 1, 2, 2, 2])
    if mode == (2, 2):
        return np.einsum("i...,i...,i->...", A, B, weights)
    elif mode == (4, 4):
        return np.einsum("ik...,kj...,k->ij...", A, B, weights)
    elif mode == (4, 2):
        return np.einsum("ij...,j...,j->i...", A, B, weights)
    elif mode == (2, 4):
        return np.einsum("i...,ij...,i->j...", A, B, weights)


def dya(A, B):
    "The dyadic (outer) product."
    return A.reshape(-1, 1, *A.shape[1:]) * B.reshape(1, -1, *B.shape[1:])


def dev(A):
    "The deviatoric part of a 3x3 tensor."
    return A - trace(A) / 3 * eye(A)


def cdya_ik(A, B):
    "The inner-crossed dyadic (outer) product."
    i, j = [a.ravel() for a in np.indices((9, 9))]

    a = np.array(
        [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)]
    )
    b = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2]).reshape(3, 3)

    i, j, k, l = np.hstack([a[i], a[j]]).T

    ik = b[i, k].reshape(3, 3, 3, 3)
    jl = b[j, l].reshape(3, 3, 3, 3)

    return A[ik] * B[jl]


def cdya_il(A, B):
    "The outer-crossed dyadic (outer) product."
    i, j = [a.ravel() for a in np.indices((9, 9))]

    a = np.array(
        [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)]
    )
    b = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2]).reshape(3, 3)

    i, j, k, l = np.hstack([a[i], a[j]]).T

    il = b[i, l].reshape(3, 3, 3, 3)
    kj = b[k, j].reshape(3, 3, 3, 3)

    return A[il] * B[kj]


def cdya(A, B):
    "The inner- and outer-crossed dyadic (outer) product."
    i, j = [a.ravel() for a in np.indices((6, 6))]

    a = np.array([(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)])
    b = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2]).reshape(3, 3)

    i, j, k, l = np.hstack([a[i], a[j]]).T

    ik = b[i, k].reshape(6, 6)
    jl = b[j, l].reshape(6, 6)

    il = b[i, l].reshape(6, 6)
    kj = b[k, j].reshape(6, 6)

    return (A[ik] * B[jl] + A[il] * B[kj]) / 2


def eigh(A, fun=None):
    "Eigenvalues and -bases of matrix A."
    wA, vA = np.linalg.eigh(astensor(A).T)
    MA = np.einsum("...ia,...ja->...ija", vA, vA)
    if fun is not None:
        wA = fun(wA)
    return wA.T, np.array([asvoigt(M) for M in MA.T])



class ThirdOrderDeformation:
    r"""Third Order Deformation isotropic hyperelastic material formulation based on the
    first and second invariant of the right Cauchy-Green deformation tensor. The strain
    energy density per unit undeformed volume is given as a sum of multivariate
    polynomials.

    ..  math::

        \psi(I_1, I_2) =
            \sum_{(i+j) \ge 1}^{(i+2j) \le 3} C_{ij}~(I_1 - 3)^i (I_2 - 3)^j

    The first partial derivatives of the strain energy density w.r.t. the
    invariants are given below.

    ..  math::

        \frac{\partial \psi}{\partial I_1} &=
            \sum_{i \ge 1} C_{ij}~i~(I_1 - 3)^{i-1} (I_2 - 3)^j

        \frac{\partial \psi}{\partial I_2} &=
            \sum_{j \ge 1} C_{ij}~(I_1 - 3)^i~j~(I_2 - 3)^{j-1}

    Furthermore, the second partial derivatives of the strain energy density w.r.t. the
    invariants are carried out.

    ..  math::

        \frac{\partial^2 \psi}{\partial I_1~\partial I_1} &=
            \sum_{i \ge 2} C_{ij}~i~(i-1)~(I_1 - 3)^{i-2} (I_2 - 3)^j

        \frac{\partial^2 \psi}{\partial I_2~\partial I_2} &=
            \sum_{j \ge 2} C_{ij}~(I_1 - 3)^i~j~(j-1)~(I_2 - 3)^{j-2}

        \frac{\partial^2 \psi}{\partial I_1~\partial I_2} &=
            \sum_{i \ge 1, j \ge 1} C_{ij}~i~(I_1 - 3)^{i-1}~j~(I_2 - 3)^{j-1}


    """

    def __init__(self, C10=0, C01=0, C11=0, C20=0, C30=0):
        """Initialize the Third Order Deformation material formulation with its
        parameters.
        """

        self.C10 = C10
        self.C01 = C01
        self.C11 = C11
        self.C20 = C20
        self.C30 = C30

    def gradient(self, I1, I2, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the invariants."""

        dWdI1 = (
            self.C10
            + self.C11 * (I2 - 3)
            + self.C20 * 2 * (I1 - 3)
            + self.C30 * 3 * (I1 - 3) ** 2
        )
        dWdI2 = self.C01 + self.C11 * (I1 - 3)

        return dWdI1, dWdI2, statevars

    def hessian(self, I1, I2, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the invariants."""

        d2WdI1I1 = self.C20 * 2 + self.C30 * 6 * (I1 - 3)
        d2WdI2I2 = 0
        d2WdI1I2 = self.C11

        return d2WdI1I1, d2WdI2I2, d2WdI1I2



from ._deformation import DeformationSpace
from ._dilatational import DilatationalSpace
from ._distortional import DistortionalSpace

__all__ = ["DistortionalSpace", "DilatationalSpace", "DeformationSpace"]
